# wsp-tools is TEM data analysis and simulation tools developed by WSP as a grad student in the McMorran Lab.
# Copyright (C) 2021  William S. Parker
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Module to generate spatial modes and calculate beam parameters.
"""

import numpy as np
from . import constants as _
from scipy.special import eval_genlaguerre, factorial

__all__ = ['energy','p','dB','k','omega','v_g','v_p','bessel','besselPacket','zR','w','roc','LG']

def energy(T_eV):
	"""Calculates the total electron energy from its kinetic energy.

	**Parameters**

	* **T_eV** : _number_ <br />
	the kinetic energy in eV.

	**Returns**

	* **energy(T_eV)** : _number_ <br />
	the total energy in eV.
	"""
	return(T_eV * _.e + _.m_e * _.c**2)

def p(T_eV):
	"""Calculates the momentum from the electron's kinetic energy.

	**Parameters**

	* **T_eV** : _number_ <br />
	the kinetic energy in eV.

	**Returns**

	* **p(T_eV)** : _number_ <br />
	the momentum in N s.
	"""
	return(1/_.c * np.sqrt(energy(T_eV)**2 - (_.m_e * _.c**2)**2))

def dB(T_eV):
	"""Calculates the de Broglie wavelength from the electron's kinetic energy.

	**Parameters**

	* **T_eV** : _number_ <br />
	the kinetic energy in eV.

	**Returns**

	* **dB(T_eV)** : _number_ <br />
	the de Broglie wavelength in m.
	"""
	return(_.h / p(T_eV))

def k(T_eV):
	"""Calculates the wavenumber from the electron's kinetic energy.

	**Parameters**

	* **T_eV** : _number_ <br />
	the kinetic energy in eV.

	**Returns**

	* **k(T_eV)** : _number_ <br />
	the wavenumber in m<sup>-1</sup>.
	"""
	return(p(T_eV) / _.hbar)

def omega(T_eV):
	"""Calculates the angular frequency from the electron's kinetic energy.

	**Parameters**

	* **T_eV** : _number_ <br />
	the kinetic energy in eV.

	**Returns**

	* **omega(T_eV)** : _number_ <br />
	the angular frequency in rad s<sup>-1</sup>.
	"""
	return(energy(T_eV) / _.hbar)

def v_g(T_eV):
	"""Calculates the group velocity from the electron's kinetic energy.

	**Parameters**

	* **T_eV** : _number_ <br />
	the kinetic energy in eV.

	**Returns**

	* **v_g(T_eV)** : _number <br />
	the group velocity in m s<sup>-1</sup>.
	"""
	return(_.c * k(T_eV) / np.sqrt(_.m_e**2 * _.c**2 / _.hbar**2 + k(T_eV)**2))

def v_p(T_eV):
	"""Calculates the phase velocity from the electron's kinetic energy.

	**Parameters**

	* **T_eV** : _number_ <br />
	the kinetic energy in eV.

	**Returns**

	* **v_p(T_eV)** : _number_ <br />
	the phase velocity in m s<sup>-1</sup>.
	"""
	return(omega(T_eV)/k(T_eV))

def bessel(x, y, z = 0, l = 0, kz0 = k(3e5),
			kperp = 0.0005*k(3e5), dkperp = 0, N = 30):
	"""Creates a Bessel beam by adding plane waves.

	The spectrum is a circle in k-space <br />`(kz0, dkperp + kperp * cos(theta), kperp * sin(theta))`.

	**Parameters**

	* **x** : _number, ndarray_ <br />
	the x-coordinates over which to calculate the beam.

	* **y** : _number, ndarray_ <br />
	the y-coordinates over which to calculate the beam.

	* **z** : _number, ndarray, optional_ <br />
	the z-coordinates over which to calculate the beam. <br />
	Default is `z = 0`.

	* **l** : _number, optional_ <br />
	the winding number, or chirality, of the beam. <br />
	Default is `l = 0`.

	* **kz0** : _number, optional_ <br />
	the z-coordinate of the spectrum. That is, the z component of the beam's wavevector. <br />
	Default is `kz0 = k(3e5)`.

	* **kperp** : _number, optional_ <br />
	the perpendicular component of the spectrum. That is, the perpendicular component of the beam's wavevector. This is the radius of the beam's spectrum. <br />
	Default is `kperp = 5e-4 * k(3e5)`.

	* **dkperp** : _number, optional_ <br />
	the perpendicular offset of the spectrum. This is applied in the `k_x` direction. <br />
	Default is `dkperp = 0`.

	* **N** : _int, optional_ <br />
	The number of plane waves to use. <br />
	Default is `N = 30`.

	**Returns**

	* **mode** : _ndarray_ <br />
	The complex amplitudes of the Bessel beam at the specified x, y, and z positions.
	Shape is broadcast between x, y, and z inputs.
	"""
	mode = x*y*z*0j
	w_0 = np.sqrt((kz0**2 + kperp**2)*_.c**2 + _.m_e**2*_.c**4/_.hbar**2)
	for n in range(N):
		phi_n = 2*np.pi*n/N
		kxn, kyn, kzn = kperp*np.cos(phi_n) + dkperp, kperp*np.sin(phi_n), kz0
		w_n = np.sqrt(w_0**2 + dkperp**2 * _.c**2 + 2*kxn*dkperp*_.c**2)
		mode += np.exp(1j*(kxn*x + kyn*y + kzn*z + l*phi_n))
	mode *= 1/np.sqrt(N)
	return(mode)

def besselPacket(t = 0, l = 0,
		kres = 2**7, kmin = -3*k(3e5), kmax = 3*k(3e5),
		kz0 = k(3e5), kperp = .5*k(3e5), dkperp = 0,
		sig = 0.05*k(3e5)):
	"""Creates a bessel beam by Fourier transforming a Gaussian spectrum.

	The spectrum is a gaussian centered on a circle in
	k-space <br /> `(k_z, dkperp + kperp * cos(theta), kperp * sin(theta))`.

	**Parameters**

	* **t** : _number, optional_ <br />
	time in seconds. <br />
	Default is `t = 0`.

	* **l** : _number, optional_ <br />
	the winding number, or chirality, of the beam. <br />
	Default is `l = 0`.

	* **kres** : _int, optional_ <br />
	the resolution of the k-space; also, the resolution of the output beam. Note that the function will be faster if the resolution is a power of 2. <br />
	Default is `kres = 128`.

	* **kmin** : _number, optional_ <br />
	the minimum value in k-space. This is applied to the x, y, and z components of the wavenumber. <br />
	Default is `kmin = -3 * k(3e5)`.

	* **kmax** : _number, optional_ <br />
	the maximum value in k-space. This is applied to the x, y, and z components of the wavenumber. <br />
	Default is `kmax = 3 * k(3e5)`.

	* **kz0** : _number, optional_ <br />
	the z-coordinate of the spectrum. That is, the z component of the beam's wavevector. <br />
	Default is `kz0 = k(3e5)`.

	* **kperp** : _number, optional_ <br />
	the perpendicular component of the spectrum. That is, the perpendicular component of the beam's wavevector. This is the radius of the beam's spectrum. <br />
	Default is `kperp = 5e-4 * k(3e5)`.

	* **dkperp** : _number, optional_ <br />
	the perpendicular offset of the spectrum. This is applied in the `k_x` direction. <br />
	Default is `dkperp = 0`.

	* **sig** : _number, optional_ <br />
	the standard deviation of the Gaussian envelope around the circle in k-space. <br />
	Default is `sig = 0.05 * k(3e5)`.

	**Returns**

	* **mode** : _ndarray_ <br />
	three-dimensional array containing the complex amplitudes of the Bessel packet. Shape is `kres`x`kres`x`kres`.
	"""
	# K = np.linspace(kmin, kmax, kres)
	kx, ky, kz = np.ogrid[kmin:kmax:1j*kres,kmin:kmax:1j*kres,kmin:kmax:1j*kres]
	w = _.c * np.sqrt((kx**2 + ky**2 + kz**2) + _.c**2 * _.m_e**2 /_.hbar**2)
	# cylindrical
	kr = np.sqrt(kx**2 + ky**2)
	mode = kz*0j
	phi = np.arctan2(ky,kx)
	spec = np.exp( -( (kr-kperp)**2 + (kz-kz0)**2 ) / (2*sig**2)) \
			* np.exp(1j*phi*l-1j*w*t)
	### Include first ifftshift because spec is [-k0, k0], not [0,k0]
	mode = np.fft.fftshift(np.fft.ifftn(np.fft.ifftshift(spec)))
	return(mode)

def zR(k, w0):
	"""Rayleigh range as a function of wavenumber and beam waist.

	**Parameters**

	* **k** : _number_ <br />
	Wavenumber.

	* **w0** : _number_ <br />
	Beam waist.

	**Returns**

	* **zR(k, w0)** : _number_ <br />
	the Rayleigh range.
	"""
	lam = 2*_.pi/k
	return(_.pi * w0**2 / lam)

def w(z,w0,k):
	"""Spot size as a function of z, beam waist, and wavenumber.

	**Parameters**

	* **z** : _number_ <br />
	z-position.

	* **w0** : _number_ <br />
	beam waist.

	* **k** : _number_ <br />
	wavenumber.

	**Returns**

	* **w(z, w0, k)** : _number_ <br />
	the spot size at z.
	"""
	return(w0 * np.sqrt(1 + (z/zR(k,w0))**2))

np.seterr(divide='ignore')
def roc(z, w0, k):
	"""Radius of curvature as a function of z, beam waist, and wavenumber.

	**Parameters**

	* **z** : _number_ <br />
	z-position.

	* **w0** : _number_ <br />
	beam waist.

	* **k** : _number_ <br />
	wavenumber.

	**Returns**

	* **roc(z, w0, k)** : _number_ <br />
	Radius of curvature.
	"""
	return(z + np.divide(zR(k,w0)**2,z))

def LG(x, y, z = 0, l = 0, p = 0, w_0 = 2e-6, lam = 1.97e-12):
	"""Generates a Laguerre-Gauss spatial mode.

	Note: broadcasting is not supported - if x and y are both 1d arrays, the
	result will be a 1darray and will NOT make sense. Likewise if x, y are 2d
	arrays, but z is a 1d array, the result will be an error.

	**Parameters**

	* **x** : _number, ndarray_ <br />
	The x-coordinates over which to calculate the beam.

	* **y** : _number, ndarray_ <br />
	The y-coordinates over which to calculate the beam.

	* **z** : _number, ndarray, optional_ <br />
	The z-coordinates over which to calculate the beam. <br />
	Default is `z = 0`.

	* **l** : _number, optional_ <br />
	The winding number, or chirality, of the beam. <br />
	Default is `l = 0`.

	* **p** : _number, optional_ <br />
	The radial index of the beam. <br />
	Default is `p = 0`.

	* **w_0** : _number, optional_ <br />
	The beam waist. <br />
	Default is `w_0 = 2e-6`.

	* **lam** : _number, optional_ <br />
	The beam's wavelength. <br />
	Default is `lam = 1.97e-12` (the relativistic wavelength of a 300keV electron).

	**Returns**

	* **mode** : _ndarray_ <br />
	The complex amplitudes of a Laguerre Gaussian beam at the specified x, y, and z positions.
	"""
	r, theta = np.sqrt(x**2 + y**2), np.arctan2(y,x)
	Clp = np.sqrt(2*factorial(p)/np.pi/factorial(p+np.abs(l)))
	z_r = np.pi*w_0**2/lam
	w_z = w_0*np.sqrt(1 + z**2/z_r**2)
	gouy = (2*p + np.abs(l) + 1) * np.nan_to_num(np.arctan2(z,z_r))
	r = np.sqrt(x**2+y**2); theta = np.arctan2(y,x)
	mode = Clp \
			* w_0/w_z \
			* (r * np.sqrt(2)/w_z)**np.abs(l) \
			* np.exp(-r**2/w_z**2) \
			* eval_genlaguerre(p,np.abs(l),(2*r**2)/w_z**2) \
			* np.exp(-1j * 2 * np.pi / lam * r**2 * z/ 2 /(z**2 + z_r**2)) \
			* np.exp(-1j * l * theta) \
			* np.exp(1j * (1 + np.abs(l) * 2*p))
	mode /= np.max(np.abs(mode))
	return(mode)



# wsp-tools is TEM data analysis and simulation tools developed by WSP as a grad student in the McMorran Lab.
# Copyright (C) 2021  William S. Parker
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from . import constants as _
import numpy as np
from deprecated import deprecated

__all__ = ['abphase2d']

@deprecated(version='1.2.0', reason='You should now use wsp_tools.ltem.ab_phase.')
def abphase2d(mx, my, mz, Lx=1e-6, Ly=1e-6, p=np.array([0,0,1]), t=60e-9):
	"""DEPRECATED. You should instead use `wsp_tools.ltem.ab_phase.`

	Calculates the Aharonov-Bohm phase acquired by an electron passing through a 2d
	magnetization.

	**Parameters**

	* **mx** : _ndarray_ <br />
	The x-component of magnetization. Should be a 2-dimensional array.

	* **my** : _ndarray_ <br />
	The y-component of magnetization. Should be a 2-dimensional array.

	* **mz** : _ndarray_ <br />
	The z-component of magnetization. Should be a 2-dimensional array.

	* **Lx** : _number, optional_ <br />
	The x-length of the area calculated. (i.e., Lx = xmax - xmin). <br />
	Default is `Lx = 1e-6`.

	* **Ly** : _number, optional_ <br />
	The y-length of the area calculated. (i.e., Ly = ymax - ymin). <br />
	Default is `Ly = 1e-6`.

	* **p** : _ndarray, optional_ <br />
	A unit vector in the direction of electron propagation. Should be a 1darray with length 3. <br />
	Default is `p = np.array([0,0,1])` (the electron is propagating in the z-direction).

	* **t** : _number, optional_ <br />
	The thickness of the 2d magnetization. <br />
	Default is `t = 60e-9`.

	**Returns**

	* **phi** : _ndarray_
	The phase acquired by an electron passing through the specified magnetization.
	Output is a 2d array of the same shape as mx, my, and mz.
	"""
	Mx = np.fft.fft2(mx)
	My = np.fft.fft2(my)
	Mz = np.fft.fft2(mz)
	M = np.array([Mx, My, Mz])
	SI = np.fft.fftfreq(M[0].shape[1],Lx/M[0].shape[1])
	SJ = np.fft.fftfreq(M[0].shape[0],Ly/M[0].shape[0])
	si, sj = np.meshgrid(SI, SJ)
	s = np.array([si,sj,0*si])
	s_mag = np.sqrt(np.einsum('ijk,ijk->jk',s,s))
	s_mag[s_mag == 0] = np.max(s_mag)/s_mag.shape[0]/1000000

	sig = np.nan_to_num(s/s_mag, nan=0,posinf=0,neginf=0)
	Gts = 1/(np.einsum('i,ijk->jk',p,sig)**2 + p[2]**2) \
			* np.sinc(t*np.einsum('i,ijk->jk',p,sig)/p[2])
	weights = np.zeros_like(M[0], dtype=complex)

	pxpxM = np.cross(p, np.cross(p, M, axisb=0))
	sigxz = np.cross(sig, np.array([0,0,1]), axisa=0)
	d = np.einsum('ijk,ijk->ij', sigxz, pxpxM)
	weights = 1j * np.nan_to_num(t/s_mag, nan=0,neginf=0,posinf=0) * Gts * d

	phi = 2*_.e/_.hbar/_.c * np.fft.ifft2(weights)
	return(phi.real)

@deprecated(version='1.2.0', reason='You should now use wsp_tools.ltem.img_from_phase')
def propagate(cphase, dx = 1, dy = 1, defocus=0, wavelength=1.97e-12):
	"""Calculates the Lorentz image given a specific phase and defocus.

	This takes into account the microscope and phase transfer functions.

	**Parameters**

	* **x** : _ndarray_ <br />
	The x coordinates of the input complex phase. Dimension should be 2.

	* **y** : _ndarray_ <br />
	The y coordinates of the input complex phase. Dimension should be 2.

	* **cphase**: _complex ndarray_ <br />
	The complex phase to propagate. Dimension should be two. May be complex
	to allow for attenuation through the sample.

	* **defocus** : _number, optional_ <br />
	The defocus at which the phase was acquired. <br />
	Default is `defocus = 0`.

	* **wavelength** : _number, optional_ <br />
	The wavelength of the electron. Scales the phase transfer function and the reciprocal coordinates. <br />
	Default is `wavelength = 1.97e-12` (the relativistic wavelength of a 300kV electron).

	* **focal_length** : _number, optional_ <br />
	The focal length of the lens, which scales the reciprocal coordinates. <br />
	Default is `focal_length = 1`.

	**Returns**

	* **psi_out** : _complex ndarray_ <br />
	The transverse complex amplitude in the image plane. Output has the same
	shape as x, y, and cphase.
	"""
	U = np.fft.fftfreq(cphase.shape[1], dx)
	V = np.fft.fftfreq(cphase.shape[0], dy)
	qx, qy = np.meshgrid(U, V)
	psi_0 = np.exp(1j * cphase)
	psi_q = np.fft.fft2(psi_0)
	psi_out = np.fft.ifft2(psi_q * T(qx, qy, defocus, wavelength))
	return(psi_out)

def T(qx, qy, defocus, wavelength):
	"""Utility function for propagate(). Microscope transfer function.
	"""
	out = aperture(qx, qy) * np.exp(-1j * chi(qx, qy, defocus, wavelength))
	return(out)

def chi(qx, qy, defocus, wavelength):
	"""Utility function for propagate(). Phase transfer function.
	"""
	return(_.pi * wavelength * defocus * (qx**2 + qy**2))

def aperture(qx, qy):
	"""Utility function for propagate(). Circular aperture.
	"""
	out = np.zeros_like(qx)
	out[np.sqrt(qx**2 + qy**2) < np.max(np.sqrt(qx**2 + qy**2))] = 1
	return(out)

@deprecated(version='1.2.0', reason='You should now use wsp_tools.ltem.jchessmodel')
def jchessmodel(x, y, z=0, **kwargs):
	"""Calculates the magnetization of a hopfion based on Jordan Chess' model.

	**Parameters**

	* **x** : _number, ndarray_ <br />
	The x-coordinates over which to calculate magnetization.

	* **y** : _number, ndarray_ <br />
	The y-coordinates over which to calculate magnetization.

	* **z** : _number, ndarray, optional_ <br />
	The z-coordinates over which to calculate magnetization. Note, if z is an
	ndarray, then x, y, and z should have the same shape rather than relying
	on array broadcasting. <br />
	Default is `z = 0`.

	* **aa**, **ba**, **ca** : _number, optional_ <br />
	In this model, the thickness of the domain wall is set by a
	Gaussian function, defined as `aa * exp(-ba * z**2) + ca`. <br />
	Defaults are `aa = 5`, `ba = 5`, `ca = 0`.

	* **ak**, **bk**, **ck** : _number, optional_ <br />
	In this model, the thickness of the core is set by a Gaussian function,
	defined as `ak * exp(-bk * z**2) + ck`. <br />
	Defaults are `ak = 5e7`, `bk = -50`, `ck = 0`.

	* **bg**, **cg** : _number, optional_ <br />
	In this model, the helicity varies as a function of z, given
	as `pi / 2 * tanh( bg * z ) + cg`. <br />
	Defaults are `bg = 5e7`, `cg = pi/2`.

	* **n** : _number, optional_ <br />
	The skyrmion number. <br />
	Default is `n = 1`.


	**Returns**

	* **mx** : _ndarray_ <br />
	The x-component of magnetization. Shape will be the same as x and y.

	* **my** : _ndarray_ <br />
	The y-component of magnetization. Shape will be the same as x and y.

	* **mz** : _ndarray_ <br />
	The z-component of magnetization. Shape will be the same as x and y.
	"""
	p = {   'aa':5, 'ba':5, 'ca':0,
			'ak':5e7, 'bk':-5e1, 'ck':0,
			'bg':5e7, 'cg':_.pi/2, 'n': 1}
	for key in kwargs.keys():
		if not key in p.keys(): return("Error: {:} is not a kwarg.".format(key))
	p.update(kwargs)

	r, phi = np.sqrt(x**2+y**2), np.arctan2(y,x)

	alpha_z = p['aa'] * np.exp(-p['ba'] * z**2) + p['ca']
	k_z = p['ak'] * np.exp(-p['bk'] * z**2) + p['ck']
	gamma_z = _.pi / 2 * np.tanh(p['bg'] * z) + p['cg']
	Theta_rz = 2 * np.arctan2((k_z * r)**alpha_z, 1)

	mx = np.cos(p['n']*phi%(2*_.pi)-gamma_z) * np.sin(Theta_rz)
	my = np.sin(p['n']*phi%(2*_.pi)-gamma_z) * np.sin(Theta_rz)
	mz = np.cos(Theta_rz)
	return(mx, my, mz)



# wsp-tools is TEM data analysis and simulation tools developed by WSP as a grad student in the McMorran Lab.
# Copyright (C) 2021  William S. Parker
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Module to generate rgba data from scalar values.
"""

import numpy as np
import matplotlib.pyplot as plt
from . import constants as _

__all__ = ['cielab_cmap','cielab_image','rgba']

def cielab_cmap(samples=256):
	"""Creates a `matplotlib.colors.ListedColormap` of the CIELAB color space.

	**Parameters**

	* **samples** : _number, optional_ <br />
	The number of samples. Any additional values will be nearest-neighbor interpolated, per matplotlib. <br />
	Default is `samples = 256`.

	**Returns**

	* **cmap** : _ListedColormap_ <br />
	A colormap, that can be used normally: `plt.imshow(data, cmap=cmap)`.
	"""
	from matplotlib.colors import ListedColormap
	angles = np.linspace(0,2*_.pi,samples)
	cvals = np.exp(1j*angles).reshape(1, samples)
	rgbavals = cielab_image(cvals).squeeze()/255
	cmap = ListedColormap(rgbavals)
	return(cmap)

def cielab_image(data, brightness = 'intensity', alpha = 'uniform'):
	"""Converts complex values to rgba data based on the CIELAB color space.

	The output color will represent the complex angle, and brightness may
	represent either intensity or amplitude.

	The CIELAB color space is intended to be perceptually uniform - none of the
	colors look brighter or darker than the others.

	**Parameters**

	* **data** : _ndarray_ <br />
	An array with the data to represent. Dtype may be complex or real - if real,
	the color will be uniform, and values will be represented by brightness.

	* **brightness** : _string, optional_ <br />
	Allowed values: `'intensity'`, `'amplitude'`, `'uniform'`. <br />
	Default is `brightness = 'intensity'`.

	* **alpha** : _string, optional_ <br />
	Allowed values: `'intensity'`, `'amplitude'`, `'uniform'`. Determines the alpha
	component of the rgba value. <br />
	Default is `alpha = 'uniform'`.

	**Returns**

	* **rgba_image_components** : _ndarray_ <br />
	The rgba components calculated from scalar values. If the input array has
	shape NxN, the output array will have shape NxNx4.
	"""
	data /= np.max(np.abs(data))
	rgba_image_components = np.zeros(np.append(data.shape, 4),dtype=np.uint8)
	if brightness == 'uniform':
		bvalue = 255
	elif brightness == 'intensity':
		bvalue = np.absolute(data)**2 * 255
	elif brightness == 'amplitude':
		bvalue = np.absolute(data) * 255
	if alpha == 'uniform':
		avalue = 255
	elif alpha == 'intensity':
		avalue = np.absolute(data)**2 * 255
	elif alpha == 'amplitude':
		avalue = np.absolute(data) * 255
	hue = (np.angle(data) + np.pi) / 2
	pi6 = np.pi/6
	def sin2(array, offset):
		return(np.sin(array-offset)**2)
	r = sin2(hue, .15*pi6) + 0.35 * sin2(hue, 3.15 * pi6)
	b = sin2(hue, 4.25 * pi6)
	g = .6*sin2(hue, 2*pi6) + 0.065 * sin2(hue, 5.05 * pi6) + 0.445*b + 0.33*r
	rgba_image_components[...,0] = (r * bvalue).astype(np.uint8)
	rgba_image_components[...,1] = (g * bvalue).astype(np.uint8)
	rgba_image_components[...,2] = (b * bvalue).astype(np.uint8)
	rgba_image_components[...,3] = np.full(data.shape, fill_value=avalue, dtype=np.uint8)
	return(rgba_image_components)

def rgba(mode, cmap = 'uniform', brightness = 'intensity', alpha = 'uniform'):
	"""Converts a 2d complex array to rgba data.

	**Parameters**

	* **mode** : _complex ndarray_ <br />
	An array with the data to represent. Dtype may be complex or real - if real,
	the color will be uniform, and values will be represented by brightness.

	* **cmap** : _string, optional_ <br />
	If `cmap = 'uniform'`, the CIELAB color space will be used. Otherwise, any
	pyplot ScalarMappable may be used. <br />
	Default is `cmap = 'uniform'`.

	* **brightness** : _string, optional_ <br />
	Allowed values: `'intensity'`, `'amplitude'`, `'uniform'`. <br />
	Default is `brightness = 'intensity'`.

	* **alpha** : _string, optional_ <br />
	Allowed values: `'intensity'`, `'amplitude'`, `'uniform'`. Determines the alpha
	component of the rgba value. <br />
	Default is `alpha = 'uniform'`.

	**Returns**

	* **rgba_image_components** : _ndarray_ <br />
	The rgba components calculated from scalar values. If the input array has
	shape NxN, the output array will have shape NxNx4.

	"""
	mode /= np.max(np.abs(mode))
	if cmap == 'uniform':
		out = cielab_image(mode, brightness, alpha)
		return(out)
	colormap = plt.cm.ScalarMappable(cmap=cmap)
	out = colormap.to_rgba(np.angle(mode))
	if alpha == 'intensity':
		out[...,-1] = np.abs(mode)**2
	elif alpha == 'amplitude':
		out[...,-1] = np.abs(mode)
	if brightness == 'intensity':
		out[...,0] *= np.abs(mode)**2
		out[...,1] *= np.abs(mode)**2
		out[...,2] *= np.abs(mode)**2
	elif brightness == 'amplitude':
		out[...,0] *= np.abs(mode)
		out[...,1] *= np.abs(mode)
		out[...,2] *= np.abs(mode)
	return(out)



# wsp-tools is TEM data analysis and simulation tools developed by WSP as a grad student in the McMorran Lab.
# Copyright (C) 2021  William S. Parker
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
wsp_tools contains utilities for TEM data analysis and presentation.

Features:

* Single Image TIE
* Lorentz simulations - phase calculations, propagation
* spatial mode implementations - LG, Bessel beams, Bessel packets
* basic image processing - high_pass, low_pass, clipping
* a matplotlib.pyplot wrapper
* an implementation of the CIELAB colorspace
* a scipy.constants (CODATA values) wrapper that allows unit scaling (i.e., using nanometers
instead of meters)
"""

import numpy as np
import matplotlib.pyplot as plt
import os, pdoc
from pathlib import Path
from deprecated import deprecated

from .beam import *
from .cielab import *
from .constants import setUnits
from .image_processing import *
from .pyplotwrapper import *
from .sitie import *
from .ltem import *

def docs(outdir = "."):
	"""Auto-generate documentation for wsp-tools in html.

	**Parameters**

	* **outdir** : _string_ <br />
	The directory to write the output documentation. <br />
	Default is "./".
	"""
	modules = ['wsp_tools']
	context = pdoc.Context()
	modules = [pdoc.Module(mod, context=context)
			   for mod in modules]
	pdoc.link_inheritance(context)

	def recursive_htmls(mod):
		yield mod.url(), mod.html()
		for submod in mod.submodules():
			yield from recursive_htmls(submod)

	for mod in modules:
		for module_url, html in recursive_htmls(mod):
			output_url = Path(outdir).expanduser().joinpath(module_url)
			if not Path(output_url).parent.exists():
				Path(output_url).parent.mkdir(parents=True)
			with open(output_url, "w+") as f:
				f.write(html)
	print("Documentation for wsp-tools written to: \n{}:".format(Path(outdir).joinpath(modules[0].url())))



# wsp-tools is TEM data analysis and simulation tools developed by WSP as a grad student in the McMorran Lab.
# Copyright (C) 2021  William S. Parker
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
from pathlib import Path
import numpy as np

__all__ = ['high_pass','low_pass','clip_data','shift_pos','outpath','ndap']

# %%
def high_pass(data, sigma = 7):
	"""Apply a high pass filter to a 2d-array.

	**Parameters**

	* **data** : _complex ndarray_ <br />

	* **sigma** : _number, optional_ <br />
	Standard deviation of the gaussian filter, measured in pixels. <br />
	Default is `sigma = 7`.

	**Returns**

	* **FFdata** : _complex ndarray_ <br />
	"""
	X = np.fft.fftfreq(data.shape[1], 1/data.shape[1])
	Y = np.fft.fftfreq(data.shape[0], 1/data.shape[0])
	x, y = np.meshgrid(X, Y)

	g = 1 - np.exp(-(x**2+y**2)/2/sigma**2)

	Fdata = np.fft.fft2(data)
	FFdata = np.fft.ifft2(g * Fdata)
	return(FFdata)

def low_pass(data, sigma = 10):
	"""Apply a low pass filter to a 2d-array.

	**Parameters**

	* **data** : _complex ndarray_ <br />

	* **sigma** : _number, optional_ <br />
	Standard deviation of the gaussian kernel (in real space), measured in pixels. <br />
	Default is `sigma = 10`.

	**Returns**

	* **FFdata** : _complex ndarray_ <br />
	"""
	X = np.linspace(-data.shape[1]/2, data.shape[1]/2, data.shape[1])
	Y = np.linspace(-data.shape[1]/2, data.shape[0]/2, data.shape[0])
	x, y = np.meshgrid(X, Y)
	g = np.exp(-(x**2 + y**2)/2/sigma**2)
	Fdata = np.fft.fft2(data)
	Fg = np.fft.fft2(np.fft.fftshift(g))
	Fg = Fg/np.max(Fg)
	FFdata = np.fft.ifft2(Fg * Fdata)
	return(FFdata)

def clip_data(data, sigma = 5):
	"""Clip data to a certain number of standard deviations from average.

	* **data** : _complex ndarray_ <br />

	* **sigma** : _number, optional_ <br />
	Number of standard deviations from average to clip to. <br />
	Default is `sigma = 5`.

	**Returns**

	* **data** : _complex ndarray_ <br />
	"""
	avg = np.mean(data)
	stdev = np.std(data)
	vmin = avg - sigma*stdev
	vmax = avg + sigma*stdev
	data[data < vmin] = vmin
	data[data > vmax] = vmax
	return(data)

def shift_pos(data):
	"""Shift data to be all greater than zero.

	**Parameters**

	* **data** : _complex ndarray_ <br />

	**Returns**

	* **data** : _complex ndarray_
	"""
	return(data - np.min(data))

def outpath(datadir, outdir, fname):
	"""A util to get the output filename.

	An example is easiest to explain:

	datadir: `/abs/path/to/data`

	fname: `/abs/path/to/data/plus/some/structure/too.dm3`

	outdir: `/where/i/want/to/write/data`

	This util will create the folder (if not exists):

	`/where/i/want/to/write/data/plus/some/structure`

	and return the filename:

	`/where/i/want/to/write/data/plus/some/structure/too`.

	**Parameters**

	* **datadir** : _string_ <br />
	The directory for the experiment's data. (abspath)

	* **outdir** : _string_ <br />
	The main directory where you want outputs to go. (abspath)

	* **fname** : _string_ <br />
	The name of the file in datadir. (abspath)

	**Returns**

	* **outname** : _string_ <br />
	The name of the file to save. (abspath)
	"""
	if not os.path.exists(outdir):
		os.makedirs(outdir)
	fname = os.path.splitext(fname)[0]
	subpath = os.path.relpath(os.path.dirname(fname), datadir)
	finoutdir = os.path.join(outdir, subpath)
	if not os.path.exists(finoutdir):
		os.makedirs(finoutdir)
	return(Path(os.path.join(finoutdir, os.path.basename(fname))))

# %%
class ndap(np.ndarray):
	"""A class that adds all the image processing methods to np.ndarray.

	The purpose of this class is just so you can write `myarray.high_pass().low_pass()` instead of `myarray = high_pass(low_pass(myarray))`.

	**Parameters**

	* **data** : _complex ndarray_ <br />
	Any type of ndarray - the methods are defined with a 2d array in mind.
	"""
	def __new__(cls, data):
		dummy = np.asarray(data, dtype=data.dtype).copy().view(cls)
		return(dummy)

	def __init__(self, data):
		self.isComplex = np.iscomplexobj(data)

	def high_pass(self, sigma = 7):
		"""Apply a high pass filter to a 2d-array.

		**Parameters**

		* **sigma** : _number, optional_ <br />
		Standard deviation of the gaussian filter, measured in pixels. <br />
		Default is `sigma = 7`.

		**Returns**

		* **FFdata** : _ndap_ <br />
		"""
		if self.isComplex:
			self[:,:] = high_pass(self, sigma)
		else:
			self[:,:] = np.real(high_pass(self, sigma))
		return(self)

	def low_pass(self, sigma = 100):
		"""Apply a low pass filter to a 2d-array.

		**Parameters**

		* **sigma** : _number, optional_ <br />
		Standard deviation of the gaussian filter, measured in pixels. <br />
		Default is `sigma = 100`.

		**Returns**

		* **FFdata** : _ndap_ <br />
		"""
		if self.isComplex:
			self[:,:] = low_pass(self, sigma)
		else:
			self[:,:] = np.real(low_pass(self, sigma))
		return(self)

	def clip_data(self, sigma = 5):
		"""Clip data to a certain number of standard deviations from average.

		* **sigma** : _number, optional_ <br />
		Number of standard deviations from average to clip to. <br />
		Default is `sigma = 5`.

		**Returns**

		* **data** : _ndap_ <br />
		"""
		self[:,:] = clip_data(self, sigma)
		return(self)

	def shift_pos(self):
		"""Shift data to be all greater than zero.

		**Returns**

		* **data** : _ndap_
		"""
		self[:,:] = shift_pos(self)
		return(self)



# wsp-tools is TEM data analysis and simulation tools developed by WSP as a grad student in the McMorran Lab.
# Copyright (C) 2021  William S. Parker
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Wrapper for the scipy.constants module that allows unit scaling.

To see all available constants, `print(wsp_tools.constants.__all__)`.

To set units:

```python
import wsp_tools as wt
wt.setUnits(meter = 1e-3) # set km as base unit for length

print(wt.constants.c) # outputs 299792.458
```

Note that all other modules should update automatically as well.
"""

import scipy.constants as sc

__all__ = ['s','m','kg','A','K','mol','cd','F','J','C','W','eV','c',
			'mu_0','eps_0','h','hbar','h_eV','hbar_eV','G','e','R','alpha',
			'N_A','k_B','k_B_eV','sigma','Wien','Rydberg','m_e','m_p',
			'm_n','pi','setUnits']

def setUnits(second=1, meter=1, kilogram=1, Amp=1, Kelvin=1, mole=1, candela=1):
	"""Sets the units across the wsp_tools module.

	i.e. setUnits(meter = 1000) sets the millimeter as the base unit for length.

	**Parameters**

	* **second** : _number, optional_ <br />
	The SI base unit for time. <br />
	Default is `second = 1`.

	* **meter** : _number, optional_ <br />
	The SI base unit for length. <br />
	Default is `meter = 1`.

	* **kilogram** : _number, optional_ <br />
	The SI base unit for mass. <br />
	Default is `kilogram = 1`.

	* **Amp** : _number, optional_ <br />
	The SI base unit for current. <br />
	Default is `Amp = 1`.

	* **Kelvin** : _number, optional_ <br />
	The SI base unit for temperature. <br />
	Default is `Kelvin = 1`.

	* **mole** : _number, optional_ <br />
	The SI base unit for amount of substance. <br />
	Default is `mole = 1`.

	* **candela** : _number, optional_ <br />
	The SI base unit for luminous intensity. <br />
	Default is `candela = 1`.
	"""
	global s,m,kg,A,K,mol,cd
	s,m,kg,A,K,mol,cd = second, meter, kilogram, Amp, Kelvin, mole, candela
	setConsts(s, m, kg, A, K, mol, cd)

def setConsts(s, m, kg, A, K, mol, cd):
	"""Utility function for `setUnits()`."""
	global F,J,C,W,eV,c,mu_0,eps_0,h,hbar,h_eV,hbar_eV,G,e,R,alpha,N_A,k_B
	global k_B_eV,sigma,Wien,Rydberg,m_e,m_p,m_n,pi
	F = s**4 * A**2 / m**2 / kg
	J = kg * m**2 / s**2
	C = A * s
	W = kg * m**2 / s**3
	eV = sc.e * J

	c       = sc.c          * m / s
	mu_0    = sc.mu_0       * m * kg / s**2 /A**2
	eps_0   = sc.epsilon_0  * F / m
	h       = sc.h          * J * s
	hbar    = sc.hbar       * J * s
	h_eV    = sc.h          * s * J / eV
	hbar_eV = sc.hbar       * s * J / eV
	G       = sc.G          * m**3 / kg / s**2
	g       = sc.g          * m / s**2
	e       = sc.e          * C
	R       = sc.R          * kg / s**2 / K / mol
	alpha   = sc.alpha
	N_A     = sc.N_A        / mol
	k_B     = sc.k          * J / K
	k_B_eV  = sc.k          * J / eV / K
	sigma   = sc.sigma      * W / m**2 / K**4
	Wien    = sc.Wien       * m * K
	Rydberg = sc.Rydberg    / m
	m_e     = sc.m_e        * kg
	m_p     = sc.m_p        * kg
	m_n     = sc.m_n        * kg
	pi      = sc.pi

setUnits()



# wsp-tools is TEM data analysis and simulation tools developed by WSP as a grad student in the McMorran Lab.
# Copyright (C) 2021  William S. Parker
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Contains utilities for reconstructing phase and magnetization from Lorentz images.

The most common use case is to generate a lorentz object from a ```.dm3``` file.
Then you can analyze using high_pass(), sitie(), clip_data(), etc.

Example:

```python
import ncempy.io.dm as dm
import wsp_tools as wt

fname = '/path/to/data.dm3'
dm3file = dm.dmReader(fname)

img = wt.lorentz(dm3file)
img.sitie(defocus=1e-3)
img.phase.clip_data(sigma=5).high_pass().low_pass()
img.Bx.clip_data(sigma=5).high_pass().low_pass()
img.By.clip_data(sigma=5).high_pass().low_pass()

img.saveMeta(outdir='someDirectory') # Saves defocus, pixelSize, etc

### plot img.Bx, img.By, img.phase, img.data, img.rawData, etc
```
"""

# %%
import numpy as np
import matplotlib.pyplot as plt
import os
from deprecated import deprecated

from . import image_processing as ip
from .pyplotwrapper import subplots
from . import constants as _
from .ltem import phase_from_img, ind_from_phase
np.seterr(divide='ignore', invalid='ignore')
import json

__all__ = ['lorentz','B_from_phase','SITIE']

@deprecated(version='1.2.0', reason='This function has not found much use and will not be maintained')
def save_lorentz(img, fname=None, fdir=''):
	"""DEPRECATED. This function will no longer be maintained.

	Saves a `lorentz` object as a `.npz` archive.

	**Parameters**

	* **img** : _lorentz_ <br />
	The lorentz object to save.

	* **fname** : _string, optional_ <br />
	If not given, the output will be the filename in the lorentz object metadata, with `.npz` rather than `.dm3`.

	* **fdir** : _string, optional_ <br />
	The directory where the lorentz object will be saved. <br />
	Default is `fdir = ''`.

	**Returns**

	* **None**
	"""
	if fname is None:
		fname = os.path.splitext(img.metadata['filename'])[0]
	if not os.path.exists(fdir):
		if not fdir == '':
			os.makedirs(fdir)
	np.savez(os.path.join(fdir, fname), **img.__dict__)
	return(None)

@deprecated(version='1.2.0', reason='This function has not found much use and will not be maintained')
def load_lorentz(fname):
	"""DEPRECATED. This function will no longer be maintained.

	Loads a `lorentz` object that has been saved as a `.npz` via `save_lorentz()`.

	**Parameters**

	* **fname** : _string_ <br />

	**Returns**

	* **img** : _lorentz_ <br />
	The saved lorentz object, converted from a `.npz`.
	"""
	f = np.load(fname, allow_pickle=True)
	fm = f['metadata'].item()
	dm3 = {'data':f['data'], 'pixelSize':[fm['pixelSize'],fm['pixelSize']], 'pixelUnit':[fm['pixelUnit'],fm['pixelUnit']], 'filename':fm['filename']}
	img = lorentz(dm3)
	img.phase = ip.ndap(f['phase'])
	img.Bx = ip.ndap(f['Bx'])
	img.By = ip.ndap(f['By'])
	img.x = f['x']
	img.y = f['y']
	fm.update(img.metadata)
	img.metadata.update(fm)
	return(img)

class lorentz:
	"""Class that contains sitie information about a lorentz image.

	**Parameters**

	* **dm3file** : _dictionary-like_ <br />
	a dm3-like file with the following keys: <br />
		<ul>
		<li> **data** : _ndarray_ <br />
		An array carrying the electron counts. </li>
		<li> **pixelSize** : _tuple_ <br />
		(_number_, _number_) - the x and y pixel sizes. </li>
		<li> **pixelUnit** : _tuple_ <br />
		(_string_, _string_) - the x and y pixel units. </li>
		<li> **filename** : _string_ <br /></li>
		</ul>
	"""
	def __init__(self, dm3file):
		self.data = ip.ndap(dm3file['data'])
		self.dx = dm3file['pixelSize'][0]
		self.dy = dm3file['pixelSize'][1]
		self.xUnit = dm3file['pixelUnit'][0]
		self.yUnit = dm3file['pixelUnit'][1]
		self.x = np.arange(0,self.data.shape[1]) * self.dx
		self.y = np.arange(0,self.data.shape[0]) * self.dy
		self.metadata = {
							'dx':float(dm3file['pixelSize'][0]),
							'dy':float(dm3file['pixelSize'][1]),
							'xUnit':dm3file['pixelUnit'][0],
							'yUnit':dm3file['pixelUnit'][1],
							'filename':dm3file['filename']
						}
		self.phase = None
		self.Bx, self.By = None, None
		self.fix_units()

	def fix_units(self, xunit=None, yunit=None):
		"""Change the pixel units to meters.

		**Parameters**

		* **unit** : _number, optional_ <br />
		The scale to multiply values by (i.e., going from 'µm' to 'm', you would use `unit = 1e-6`). If none is given, `fix_units` will try to convert from `self.pixelUnit` to meters.

		**Returns**

		* **self** : _lorentz_
		"""
		if xunit is None:
			if self.xUnit == 'nm':
				xunit = 1e-9
			elif self.xUnit == 'mm':
				xunit = 1e-3
			elif self.xUnit == 'µm':
				xunit = 1e-6
			elif self.xUnit == 'm':
				xunit = 1
		if yunit is None:
			if self.yUnit == 'nm':
				yunit = 1e-9
			elif self.yUnit == 'mm':
				yunit = 1e-3
			elif self.yUnit == 'µm':
				yunit = 1e-6
			elif self.yUnit == 'm':
				yunit = 1
		self.dx *= xunit
		self.dy *= yunit
		self.xUnit = 'm'
		self.yUnit = 'm'
		self.x *= xunit
		self.y *= yunit
		self.metadata.update({
			'dx': float(self.dx),
			'dy': float(self.dy),
			'xUnit': self.xUnit,
			'yUnit': self.yUnit
		})
		return(None)

	def sitie(self, defocus = 1, thickness = 60e-9, wavelength=1.97e-12):
		"""Carries out phase and B-field reconstruction.

		Assigns phase, Bx, and By attributes.

		Updates metadata with the defocus and wavelength.

		**Parameters**

		* **defocus** : _number, optional_ <br />
		The defocus at which the images were taken. <br />
		Default is `defocus = 1`.

		* **wavelength** : _number, optional_ <br />
		The electron wavelength. <br />
		Default is `wavelength = 1.96e-12` (relativistic wavelength of a 300kV electron).

		**Returns**

		* **self** : _lorentz_
		"""
		self.metadata.update({'defocus': defocus, 'wavelength': wavelength, 'thickness': thickness})
		self.phase = ip.ndap(phase_from_img(self.data, defocus, self.dx, self.dy, wavelength))
		self.Bx, self.By = [ip.ndap(arr) for arr in ind_from_phase(self.phase, thickness)]
		return(None)

	def preview(self, window=None):
		"""Preview the image.

		Note that unlike `pyplotwrapper`, window is in units of pixels.

		**Parameters**

		* **window** : _array-like, optional_ <br />
		Format is `window = (xmin, xmax, ymin, ymax)`. <br />
		Default is `window = (0, -1, 0, -1)`
		"""
		fig, ax = subplots(11)
		if not window is None:
			ax[0,0].setWindow(window)
		data = ip.clip_data(self.data, sigma=10)
		ax[0,0].setAxes(self.x, self.y)
		ax[0,0].set_xlabel("x ({:})".format(self.dx))
		ax[0,0].set_ylabel("y ({:})".format(self.dy))
		ax[0,0].imshow(data)
		plt.show()

	def saveMeta(self, outdir='', outname='metadata.json'):
		"""Save the metadata of the lorentz object to a file.

		**Parameters**

		* **outdir** : _string, optional_ <br />
		The directory where you'd like to save the metadata. <br />
		Default is `outdir = ''`.

		* **outname** : _string, optional_ <br />
		The name of the metadata file. <br />
		Default is `outname = 'metadata.json'`.
		"""
		if not os.path.exists(outdir):
			os.makedirs(outdir)
		with open(os.path.join(outdir, outname), 'w') as fp:
			json.dump(self.metadata, fp, indent="")

################################## SITIE #######################################
@deprecated(version='1.2.0', reason='You should now use wsp_tools.ltem.ind_from_phase. As of this version, B_from_phase just calls ind_from_phase.')
def B_from_phase(phase, thickness=1):
	"""DEPRECATED. You should instead use `wsp_tools.ltem.ind_from_phase`.

	Reconstructs the B-field from the phase profile.

	**Parameters**

	* **phase** : _ndarray_ <br />
	a 2d array representing the electron's phase.

	* **thickness** : _number_ <br />
	the thickness of the sample. <br />
	Default is `thickness = 1`.

	**Returns**

	* **Bx** : _ndarray_ <br />
	The x-component of the magnetic field.

	* **By** : _ndarray_ <br />
	The y-component of the magnetic field.
	"""
	return(ind_from_phase(phase, thickness))

@deprecated(version='1.2.0', reason='You should now use wsp_tools.ltem.phase_from_img. As of this version, SITIE just calls phase_from_img.')
def SITIE(image, defocus, pixel_size = 1, wavelength=1.97e-12):
	"""DEPRECATED. You should instead use `wsp_tools.ltem.phase_from_img`.

	Reconstruct the phase from a defocussed image.

	**Parameters**

	* **image** : _ndarray_ <br />
	the 2d image data. <br />

	* **defocus** : _number_ <br />

	* **pixel_size** : _number, optional_ <br />
	Default is `pixel_size = 1`.

	* **wavelength** : _number, optional_ <br />
	Default is `wavelength = 1.97e-12` (the relativistic wavelength of a 300kV electron).

	**Returns**

	* **phase** : _ndarray_ <br />
	"""
	return(phase_from_img(img, defocus, pixel_size, pixel_size, wavelength))



# wsp-tools is TEM data analysis and simulation tools developed by WSP as a grad student in the McMorran Lab.
# Copyright (C) 2021  William S. Parker
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from . import constants as _
import numpy as np
from .image_processing import shift_pos

__all__ = [
			'ab_phase',
			'B_from_mag',
			'A_from_mag',
			'img_from_mag',
			'img_from_phase',
			'ind_from_phase',
			'ind_from_img',
			'phase_from_img',
			'jchessmodel',
			'ind_from_mag']

def ind_from_mag(mx, my, mz, dx=1, dy=1, thickness=60e-9, p = np.array([0,0,1])):
	"""Calculate the magnetic induction imparted on a fast electron by a magnetized sample.

	This is shorthand for `ind_from_phase(ab_phase(*args))`.

	**Parameters**

	* **mx** : _ndarray_ <br />
	The x-component of magnetization. Should be two or three dimensions.

	* **my** : _ndarray_ <br />
	The y-component of magnetization. Should be two or three dimensions.

	* **mz** : _ndarray_ <br />
	The z-component of magnetization. Should be two or three dimensions.

	* **dx** : _number, optional_ <br />
	The spacing between pixels/samples in mx, my, mz, in the x-direction. <br />
	Default is `dx = 1`.

	* **dy** : _number, optional_ <br />
	The spacing between pixels/samples in mx, my, mz, in the y-direction. <br />
	Default is `dy = 1`.

	* **thickness** : _number, optional_ <br />
	The thickness of each slice of the x-y plane (if no z-dependence, the thickness of the sample). <br />
	Default is `thickness = 60e-9`.

	* **p** : _ndarray, optional_ <br />
	A unit vector representing the direction of the electron's path. Shape should be (3,). <br />
	Default is `p = np.array([0,0,1])`.

	**Returns**

	* **Bx** : _ndarray_ <br />
	The x-component of the magnetic induction.

	* **By** : _ndarray_ <br />
	The y-component of the magnetic induction.
	"""
	phase = ab_phase(mx, my, mz, dx, dy, thickness, p)
	if len(phase.shape) > 2:
		phase = np.sum(phase, axis=-1)
	Bx, By = ind_from_phase(phase, thickness)
	return(Bx, By)

def ab_phase(mx, my, mz, dx=1, dy=1, thickness=60e-9, p = np.array([0,0,1])):
	"""Calculate the Aharanov-Bohm phase imparted on a fast electron by a magnetized sample.

	This is an implementation of the algorithm described in Mansuripur et. al, _Computation of electron diffraction patterns in Lorentz TEM_, Eq (13) specifically.

	The shape of the output array is the same as mx, my, and mz. If mx, my, mz are three dimensional (i.e., have z-dependence as well as x and y), the third dimension of the output array represents the Aharanov-Bohm phase of each z-slice of the magnetization.

	**Parameters**

	* **mx** : _ndarray_ <br />
	The x-component of magnetization. Should be two or three dimensions.

	* **my** : _ndarray_ <br />
	The y-component of magnetization. Should be two or three dimensions.

	* **mz** : _ndarray_ <br />
	The z-component of magnetization. Should be two or three dimensions.

	* **dx** : _number, optional_ <br />
	The spacing between pixels/samples in mx, my, mz, in the x-direction. <br />
	Default is `dx = 1`.

	* **dy** : _number, optional_ <br />
	The spacing between pixels/samples in mx, my, mz, in the y-direction. <br />
	Default is `dy = 1`.

	* **thickness** : _number, optional_ <br />
	The thickness of each slice of the x-y plane (if no z-dependence, the thickness of the sample). <br />
	Default is `thickness = 60e-9`.

	* **p** : _ndarray, optional_ <br />
	A unit vector representing the direction of the electron's path. Shape should be (3,). <br />
	Default is `p = np.array([0,0,1])`.

	**Returns**

	* **phase** : _ndarray_ <br />
	The Aharanov-Bohm phase imparted on the electron by the magnetization. Shape will be the same as mx, my, mz.
	"""
	M, s, s_mag, sig, z_hat = common_stuff(mx, my, mz, dx, dy)
	Gp = G(p, sig, z_hat, thickness * s_mag)
	sig_x_z = np.cross(sig, z_hat, axisa=0, axisb=0, axisc=0)
	p_x_p_M = np.cross(p, np.cross(p, M, axisa=0, axisb=0, axisc=0), axis=0, axisb=0, axisc=0)
	weights = 2 * _.e / _.hbar / _.c * 1j * thickness / s_mag * Gp * np.einsum('i...,i...->...',sig_x_z, p_x_p_M)
	weights[:,0,0,:] = 0
	phase = np.fft.ifft2(weights, axes=(1,2), norm = 'forward')
	return(np.squeeze(phase.real))

def B_from_mag(mx, my, mz, z = 0, dx = 1, dy = 1, thickness = 60e-9):
	"""Calculate the magnetic field of a specified 2-d magnetic configuration.

	This is an adaptation of the algorithm described in Mansuripur et. al, _Computation of electron diffraction patterns in Lorentz TEM_, Eq (13) specifically.

	The output shape is (3, mx.shape[0], mx.shape[1], z.shape[0]).

	**Parameters**

	* **mx** : _ndarray_ <br />
	The x-component of magnetization. Must be a 2-d array.

	* **my** : _ndarray_ <br />
	The y-component of magnetization. Must be a 2-d array.

	* **mz** : _ndarray_ <br />
	The z-component of magnetization. Must be a 2-d array.

	* **z** : _number, optional_, ndarray_ <br />
	The z-coordinates at which to calculate the B-field. Can be a number, or a 1d-array. <br />
	Default is `z = 0`.

	* **dx** : _number, optional_ <br />
	The spacing between pixels/samples in mx, my, mz, in the x-direction. <br />
	Default is `dx = 1`.

	* **dy** : _number, optional_ <br />
	The spacing between pixels/samples in mx, my, mz, in the y-direction. <br />
	Default is `dy = 1`.

	* **thickness** : _number, optional_ <br />
	The thickness of the sample. <br />
	Default is `thickness = 60e-9`.

	**Returns**

	* **B** : _ndarray_ <br />
	The magnetic field resulting from the given 2d magnetization.
	"""
	z = np.atleast_1d(z)
	selz_m = z < - thickness / 2
	selz_z = np.abs(z) <= thickness / 2
	selz_p = z > thickness / 2
	z = z[np.newaxis,np.newaxis,np.newaxis,...]

	M, s, s_mag, sig, z_hat = common_stuff(mx, my, mz, dx, dy)
	sigp = sig + 1j * z_hat
	sigm = sig - 1j * z_hat

	A_mn = A_mn_components(
			mx.shape[0], mx.shape[1], z.shape[-1], selz_m, selz_z,
			selz_p, s_mag, z, sigm, sigp, sig, M, thickness, z_hat)
	dxA_mn = 1j * 2 * _.pi * s[0][np.newaxis, ...] * A_mn
	dyA_mn = 1j * 2 * _.pi * s[1][np.newaxis, ...] * A_mn
	dzA_mn = np.zeros((3, mx.shape[0], mx.shape[1], z.shape[-1]), dtype=complex)
	dzA_mn[...,selz_m] = 2 * _.pi * s_mag * A_mn[...,selz_m]
	dzA_mn[...,selz_p] = -2 * _.pi * s_mag * A_mn[...,selz_p]
	dzA_mn[...,selz_z] = (2 * 1j / s_mag * np.cross((
						sig
						- 0.5 * 2 * _.pi * s_mag * np.exp(2 * _.pi * s_mag * (z[...,selz_z] - thickness / 2)) * sigm
						+ 0.5 * 2 * _.pi * s_mag * np.exp(-2 * _.pi * s_mag * (z[...,selz_z] + thickness / 2)) * sigp
						), M, axisa=0, axisb=0, axisc=0))
	dzA_mn[:,0,0,selz_m] = 0
	dzA_mn[:,0,0,selz_p] = 0
	dzA_mn[:,0,0,selz_z] = -4 * _.pi * np.cross(z_hat[:,0,0,:], M[:,0,0,:], axisa=0, axisb=0, axisc=0)
	B_mn = np.zeros((3, mx.shape[0], mx.shape[1], z.shape[-1]), dtype=complex)
	B_mn[0] = dyA_mn[2] - dzA_mn[1]
	B_mn[1] = dzA_mn[0] - dxA_mn[2]
	B_mn[2] = dxA_mn[1] - dyA_mn[0]
	B = np.fft.ifft2(B_mn, axes=(1,2), norm='forward')
	return(np.squeeze(B.real))

def A_from_mag(mx, my, mz, z = 0, dx = 1, dy = 1, thickness = 60e-9):
	"""Calculate the magnetic vector potential of a specified 2-d magnetic configuration.

	This is an adaptation of the algorithm described in Mansuripur et. al, _Computation of electron diffraction patterns in Lorentz TEM_, Eq (13) specifically.

	The output shape is (3, mx.shape[0], mx.shape[1], z.shape[0]).

	**Parameters**

	* **mx** : _ndarray_ <br />
	The x-component of magnetization. Must be a 2-d array.

	* **my** : _ndarray_ <br />
	The y-component of magnetization. Must be a 2-d array.

	* **mz** : _ndarray_ <br />
	The z-component of magnetization. Must be a 2-d array.

	* **z** : _number, optional_, ndarray_ <br />
	The z-coordinates at which to calculate the B-field. Can be a number_, or a 1d-array. <br />
	Default is `z = 0`.

	* **dx** : _number, optional_ <br />
	The spacing between pixels/samples in mx, my, mz, in the x-direction. <br />
	Default is `dx = 1`.

	* **dy** : _number, optional_ <br />
	The spacing between pixels/samples in mx, my, mz, in the y-direction. <br />
	Default is `dy = 1`.

	* **thickness** : _number, optional_ <br />
	The thickness of the sample. <br />
	Default is `thickness = 60e-9`.

	**Returns**

	* **B** : _ndarray_ <br />
	The magnetic vector potential resulting from the given 2d magnetization.
	"""
	z = np.atleast_1d(z)
	selz_m = z < -thickness/2
	selz_z = np.abs(z) <= thickness/2
	selz_p = z > thickness/2
	z = z[np.newaxis,np.newaxis,np.newaxis,...]

	M, s, s_mag, sig, z_hat = common_stuff(mx, my, mz, dx, dy)
	sigp = sig + 1j * z_hat
	sigm = sig - 1j * z_hat

	A_mn = A_mn_components(
				mx.shape[0], mx.shape[1], z.shape[-1], selz_m, selz_z,
				selz_p, s_mag, z, sigm, sigp, sig, M, thickness, z_hat)

	A = np.fft.ifft2(A_mn, axes=(1,2), norm='forward')
	return(np.squeeze(A.real))

def img_from_mag(mx, my, mz, dx = 1, dy = 1, defocus = 0, thickness = 60e-9, wavelength = 1.97e-12, p = np.array([0,0,1])):
	"""Calculate the theoretical Lorentz TEM image from a given (2 or 3 dim) magnetization.

	This is an implementation of the SITIE equation (Eq 10) from J. Chess et al. 2017, _Streamlined approach to mapping the magnetic induction of skyrmionic materials_, in combination with Mansuripur et al.

	**Parameters**

	* **mx** : _ndarray_ <br />
	The x-component of magnetization. Should be two or three dimensions.

	* **my** : _ndarray_ <br />
	The y-component of magnetization. Should be two or three dimensions.

	* **mz** : _ndarray_ <br />
	The z-component of magnetization. Should be two or three dimensions.

	* **dx** : _number, optional_ <br />
	The spacing between pixels/samples in mx, my, mz, in the x-direction. <br />
	Default is `dx = 1`.

	* **dy** : _number, optional_ <br />
	The spacing between pixels/samples in mx, my, mz, in the y-direction. <br />
	Default is `dx = 1`.

	* **defocus** : _number, optional_ <br />
	The defocus - note that this should be non-zero in order to see any contrast. <br />
	Default is `defocus = 0`.

	* **wavelength** : _number, optional_ <br />
	The relativistic electron wavelength. <br />
	Default is `wavelength = 1.97e-12`.

	**Returns**

	* **img** : _ndarray_ <br />
	The intensity of the image plane.
	"""
	M, s, s_mag, sig, z_hat = common_stuff(mx, my, mz, dx, dy)
	Gp = G(p, sig, z_hat, thickness * s_mag)
	sig_x_z = np.cross(sig, z_hat, axisa=0, axisb=0, axisc=0)
	p_x_p_M = np.cross(p, np.cross(p, M, axisa=0, axisb=0, axisc=0), axis=0, axisb=0, axisc=0)
	### weights is the only thing different from ab_phase (\nabla^2 = - 4 pi^2 s_mag^2)
	weights = - 4 * _.pi **2 * 2 * _.e / _.hbar / _.c * 1j * thickness * s_mag * Gp * np.einsum('i...,i...->...',sig_x_z, p_x_p_M)
	weights[:,0,0,:] = 0

	out = wavelength * defocus / 2 / _.pi * np.fft.ifft2(weights, axes=(1,2), norm = 'forward')
	if out.shape[-1] > 1:
		out = np.sum(out, axis=-1)
	out = 1 - np.squeeze(out)
	return(out.real)

def img_from_phase(phase, dx = 1, dy = 1, defocus = 0, wavelength = 1.97e-12):
	"""Calculate the Lorentz TEM image given a two-dimensional phase distribution and defocus.

	This is an implementation of the SITIE equation (Eq 10) from J. Chess et al. 2017, _Streamlined approach to mapping the magnetic induction of skyrmionic materials_.

	**Parameters**

	* **phase** : _ndarray_ <br />
	A 2d array containing the phase of the electron at the sample plane.

	* **dx** : _number, optional_ <br />
	The spacing between pixels/samples in mx, my, mz, in the x-direction. <br />
	Default is `dx = 1`.

	* **dy** : _number, optional_ <br />
	The spacing between pixels/samples in mx, my, mz, in the y-direction. <br />
	Default is `dx = 1`.

	* **defocus** : _number, optional_ <br />
	The defocus - note that this should be non-zero in order to see any contrast. <br />
	Default is `defocus = 0`.

	* **wavelength** : _number, optional_ <br />
	The relativistic electron wavelength. <br />
	Default is `wavelength = 1.97e-12`.

	**Returns**

	* **img** : _ndarray_ <br />
	The intensity of the image plane.
	"""
	Sx = np.fft.fftfreq(phase.shape[1], dx)
	Sy = np.fft.fftfreq(phase.shape[0], dy)
	sx, sy = np.meshgrid(Sx, Sy) ### (y-dim, x-dim, z-dim)
	phase = - 4 * _.pi**2 * (sx**2 + sy**2) * np.fft.fft2(phase, norm = 'forward')
	img = 1 - wavelength * defocus / 2 / _.pi * np.fft.ifft2(phase, norm = 'forward')
	return(img.real)

def ind_from_phase(phase, thickness = 60e-9):
	"""Calculate the magnetic induction given the Aharanov-Bohm phase shift.

	This is an implementation of Eq (11) from J. Chess et al., 2017, _Streamlined approach to mapping the magnetic induction of skyrmionic materials_.

	**Parameters**

	* **phase** : _ndarray_ <br />
	A 2d array containing the electron phase immediately after the sample.

	* **thickness** : _number, optional_ <br />
	The thickness of the sample. <br />
	Default is `thickness = 60e-9`.

	**Returns**

	* **Bx** : _ndarray_ <br />
	The x-component of the magnetic induction.

	* **By** : _ndarray_ <br />
	The y-component of the magnetic induction.
	"""
	dpdy, dpdx = np.gradient(phase)
	Bx = _.hbar/_.e/thickness * dpdy
	By = -_.hbar/_.e/thickness * dpdx
	return(Bx.real, By.real)

def phase_from_img(img, defocus = 0, dx = 1, dy = 1, wavelength = 1.97e-12):
	"""Reconstruct the Aharanov-Bohm phase from a defocussed image.

	This is an implementation of the SITIE equation (Eq 10) from J. Chess et al., 2017, _Streamlined approach to mapping the magnetic induction of skyrmionic materials_.

	**Parameters**

	* **img** : _ndarray_ <br />
	The 2d image data.

	* **defocus** : _number, optional_ <br />
	Default is `defocus = 0`.

	* **dx** : _number, optional_ <br />
	The pixel spacing in the x-direction. <br />
	Default is `dx = 1`

	* **dy** : _number, optional_ <br />
	The pixel spacing in the y-direction. <br />
	Default is `dy = 1`.

	* **wavelength** : _number, optional_ <br />
	The relativistic electron wavelength. <br />
	Default is `wavelength = 1.97e-9`.

	**Returns**

	* **phase** : _ndarray_ <br />
	A 2d array containing the reconstructed Aharanov-Bohm phase shift.
	"""
	Sx = np.fft.fftfreq(img.shape[1], dx)
	Sy = np.fft.fftfreq(img.shape[0], dy)
	sx, sy = np.meshgrid(Sx, Sy)
	rhs = np.nan_to_num(2 * _.pi / wavelength / defocus * (1 - img / np.mean(img)), posinf = 0, neginf = 0)
	rhs = shift_pos(rhs)
	rhs = np.fft.fft2(rhs, norm = 'forward') / -4 / _.pi**2 / (sx**2 + sy**2)
	rhs = np.nan_to_num(rhs, posinf = 0, neginf = 0)
	phase = np.fft.ifft2(rhs, norm = 'forward')
	return(phase.real)

def ind_from_img(img, defocus = 0, dx = 1, dy = 1, thickness = 60e-9, wavelength = 1.97e-12):
	"""Reconstruct the magnetic induction from a defocussed image.

	This is an implementation of the SITIE equation Eq (10) and Eq (11) from J. Chess et al., 2017, _Streamlined approach to mapping the magnetic induction of skyrmionic materials_.

	**Parameters**

	* **img** : _ndarray_ <br />
	The 2d image data.

	* **defocus** : _number, optional_ <br />
	Default is `defocus = 0`.

	* **dx** : _number, optional_ <br />
	The pixel spacing in the x-direction. <br />
	Default is `dx = 1`

	* **dy** : _number, optional_ <br />
	The pixel spacing in the y-direction. <br />
	Default is `dy = 1`.

	* **thickness** : _number, optional_ <br />
	The thickness of the sample. <br />
	Default is `thickness = 60e-9`.

	* **wavelength** : _number, optional_ <br />
	The relativistic electron wavelength. <br />
	Default is `wavelength = 1.97e-9`.

	**Returns**

	* **Bx** : _ndarray_ <br />
	A 2d array containing the reconstructed x-component of magnetic induction.

	* **By** : _ndarray_ <br />
	A 2d array containing the reconstructed y-component of magnetic induction.
	"""
	phase = phase_from_img(img, defocus, dx, dy, wavelength)
	Bx, By = ind_from_phase(phase, thickness)
	return(Bx.real, By.real)

def jchessmodel(x, y, z=0, **kwargs):
	"""Calculates the magnetization of a hopfion based on Jordan Chess' model.

	**Parameters**

	* **x** : _number, ndarray_ <br />
	The x-coordinates over which to calculate magnetization.

	* **y** : _number, ndarray_ <br />
	The y-coordinates over which to calculate magnetization.

	* **z** : _number, ndarray, optional_ <br />
	The z-coordinates over which to calculate magnetization. Note, if z is an
	ndarray, then x, y, and z should have the same shape rather than relying
	on array broadcasting. <br />
	Default is `z = 0`.

	* **aa**, **ba**, **ca** : _number, optional_ <br />
	In this model, the thickness of the domain wall is set by a
	Gaussian function, defined as `aa * exp(-ba * z**2) + ca`. <br />
	Defaults are `aa = 5`, `ba = 5`, `ca = 0`.

	* **ak**, **bk**, **ck** : _number, optional_ <br />
	In this model, the thickness of the core is set by a Gaussian function,
	defined as `ak * exp(-bk * z**2) + ck`. <br />
	Defaults are `ak = 5e7`, `bk = -50`, `ck = 0`.

	* **bg**, **cg** : _number, optional_ <br />
	In this model, the helicity varies as a function of z, given
	as `pi / 2 * tanh( bg * z ) + cg`. <br />
	Defaults are `bg = 5e7`, `cg = pi/2`.

	* **n** : _number, optional_ <br />
	The skyrmion number. <br />
	Default is `n = 1`.

	**Returns**

	* **mx** : _ndarray_ <br />
	The x-component of magnetization. Shape will be the same as x and y.

	* **my** : _ndarray_ <br />
	The y-component of magnetization. Shape will be the same as x and y.

	* **mz** : _ndarray_ <br />
	The z-component of magnetization. Shape will be the same as x and y.
	"""
	p = {   'aa':5, 'ba':5, 'ca':0,
			'ak':5e7, 'bk':-5e1, 'ck':0,
			'bg':5e7, 'cg':_.pi/2, 'n': 1}
	for key in kwargs.keys():
		if not key in p.keys(): return("Error: {:} is not a kwarg.".format(key))
	p.update(kwargs)

	r, phi = np.sqrt(x**2+y**2), np.arctan2(y,x)

	alpha_z = p['aa'] * np.exp(-p['ba'] * z**2) + p['ca']
	k_z = p['ak'] * np.exp(-p['bk'] * z**2) + p['ck']
	gamma_z = _.pi / 2 * np.tanh(p['bg'] * z) + p['cg']
	Theta_rz = 2 * np.arctan2((k_z * r)**alpha_z, 1)

	mx = np.cos(p['n']*phi%(2*_.pi)-gamma_z) * np.sin(Theta_rz)
	my = np.sin(p['n']*phi%(2*_.pi)-gamma_z) * np.sin(Theta_rz)
	mz = np.cos(Theta_rz)
	return(mx, my, mz)

######## Helper functions

def G(p, sig, z_hat, ts_mag):
	sum1 = np.einsum('i,i...->...', p, sig)
	sum2 = np.einsum('i,i...->...', p, z_hat)
	out = 1 / (sum1**2 + sum2**2)[np.newaxis,...]
	out *= np.sinc(ts_mag * sum1 / sum2)
	return(out)

def common_stuff(mx, my, mz, dx, dy):
	#### Used for everything in LTEM sims
	#### All outputs have shape (3, y-dim, x-dim, z-dim)
	#### They just need these four dimensions so that they're broadcastable
	#### mx, my, mz either 2 or 3 dim

	mx = np.atleast_3d(mx) ### (y-dim, x-dim, z-dim)
	my = np.atleast_3d(my)
	mz = np.atleast_3d(mz)
	Mx = np.fft.fft2(mx, axes=(0,1), norm = 'forward')
	My = np.fft.fft2(my, axes=(0,1), norm = 'forward')
	Mz = np.fft.fft2(mz, axes=(0,1), norm = 'forward')
	M = np.array([Mx, My, Mz]) ### (vec, y-dim, x-dim, z-dim)

	Sx = np.fft.fftfreq(mx.shape[1], dx)
	Sy = np.fft.fftfreq(mx.shape[0], dy)
	sx, sy = np.meshgrid(Sx, Sy) ### (y-dim, x-dim)

	s = np.array([sx, sy, 0*sy])[...,np.newaxis] ### (vec, y-dim, x-dim, z-dim)
	s_mag = np.sqrt(np.einsum('i...,i...->...',s,s))[np.newaxis,...] ### (vec, y-dim, x-dim, z-dim)
	sig = s/s_mag

	z_hat = np.array([np.zeros_like(mx), np.zeros_like(mx), np.ones_like(mx)]) ### (vec, y-dim, x-dim, z-dim)
	return(M, s, s_mag, sig, z_hat)

def A_mn_components(
		xshape, yshape, zshape, selz_m, selz_z,
		selz_p, s_mag, z, sigm, sigp, sig, M, thickness, z_hat):
	### set everything to be broadcastable so we can write the A_mn equations
	### shape is: (3 vector components, y-axis, x-axis, z-axis) (thank meshgrid for switching x and y)

	A_mn = np.zeros((3, xshape, yshape, zshape), dtype=complex)
	A_mn[...,selz_m] = (2 * 1j / s_mag
						* np.exp(2 * _.pi * s_mag * z[...,selz_m])
						* np.sinh(_.pi * thickness * s_mag)
						* np.cross(sigm, M, axisa=0, axisb=0, axisc=0))
	A_mn[...,selz_z] = (2 * 1j / s_mag * np.cross((
						sig
						- 0.5 * np.exp(2 * _.pi * s_mag * (z[...,selz_z] - thickness / 2)) * sigm
						- 0.5 * np.exp(-2 * _.pi * s_mag * (z[...,selz_z] + thickness / 2)) * sigp
						), M, axisa=0, axisb=0, axisc=0))
	A_mn[...,selz_p] = (2 * 1j / s_mag
						* np.exp(-2 * _.pi * s_mag * z[...,selz_p])
						* np.sinh(_.pi * thickness * s_mag)
						* np.cross(sigp, M, axisa=0, axisb=0, axisc=0)
						)
	zero_comp = np.cross(z_hat[:,0,0,:], M[:,0,0,:], axisa=0, axisb=0, axisc=0)
	A_mn[:,0,0,selz_z] = -4 * _.pi * z[:,0,0,selz_z] * zero_comp
	A_mn[:,0,0,selz_m] = 2 * _.pi * thickness * zero_comp
	A_mn[:,0,0,selz_p] = -2 * _.pi * thickness * zero_comp
	return(A_mn)



from setuptools import setup, find_packages

setup(
	name = 'wsp_tools',
	packages = find_packages(),
	version = '1.3.1',
	author = 'William S. Parker',
	author_email = 'wparker4@uoregon.edu',
	description = 'Utilities for TEM data analysis and simulation.',
	url = 'https://github.com/McMorranLab/wsp_tools',
	project_urls={
        "Bug Tracker": "https://github.com/McMorranLab/wsp_tools/issues",
    },
	classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
	long_description = open('README.md').read(),
	long_description_content_type = "text/markdown",
	python_requires='>=3.6',
	install_requires=['numpy','matplotlib','scipy','pdoc3','deprecated']
)



# wsp-tools is TEM data analysis and simulation tools developed by WSP as a grad student in the McMorran Lab.
# Copyright (C) 2021  William S. Parker
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""A wrapper for matplotlib.pyplot containing common plotting routines.

Example

```python
import numpy as np
import wsp_tools as wt

X = np.linspace(-10,10,100)
Y = np.linspace(-10,10,100)
x, y = np.meshgrid(X, Y)
data = x + 1j*y
window = (-4, 4, -6, 6)

fig, ax = wt.subplots(12)
ax1, ax2 = ax[0,0], ax[0,1]
ax1.setAxes(X, Y)
ax1.imshow(np.abs(data))
ax1.inset(window=window)
ax2.setAxes(X, Y, window=window)
ax2.rgba(data)
wt.plt.show()
```
"""

import matplotlib.pyplot as plt
import numpy as np
from .cielab import rgba, cielab_cmap

__all__ = ['singleAx','subplots']

class singleAx():
	"""An extension of the `matplotlib.axes.Axes` class.

	This class adds macros for 2d plotting that I commonly use. In particular,
	it's easy to select only a window of your data to show, to add x-y axes,
	to add an inset, and to show the rgba version of a complex 2d array.

	Typical usage:

	```python
	X = np.linspace(-10,10,xres)
	Y = np.linspace(-10,10,yres)
	x, y = np.meshgrid(X, Y)
	data = x+1j*y
	window = [-3,7,1,4]

	fig, ax = plt.subplots()
	myax = wsp_tools.singleAx(ax)
	ax.setAxes(x, y, window)
	ax.set_xytitle('x','y','title')
	ax.rgba(data)
	plt.show()
	```

	More commonly, this class is returned by ```wsp_tools.pyplotwrapper.subplots```.

	**Parameters**

	* **ax** : _matplotlib.axes.Axes_ <br />
	"""
	def __init__(self, ax, title='', xlabel='', ylabel=''):
		self.ax = ax
		self.hasAxes = False
		self.hasWindow = False
		self.set_xytitle(xlabel, ylabel, title)

	def prePlot(self, data, step=1):
		"""Utility function that applies the axes and window before plotting.

		If you want to use a plotting function from matplotlib, you can use this
		function to get the windowed data:

		```python
		fig, axis = plt.subplots()
		ax = singleAx(axis)
		ax.setWindow(window)
		x_windowed, y_windowed, data_windowed = ax.prePlot(data)
		ax.ax.SomeOtherMatplotlibPlottingRoutine(x_windowed, y_windowed, data_windowed)
		plt.show()
		```

		**Parameters** :

		* **data** : _complex ndarray_ <br />
		The data to plot. Must be 2-dimensional.

		* **step** : _int_ <br />
		data will be returned as `data[::step,::step]` - particularly useful for
		quiver plots. <br />
		Default is `step = 1`.

		**Returns**

		* **xout** : _ndarray_ <br />
		A 1darray with x-coordinates - either the array set by `setAxes()`,
		or an array of length `data.shape[1]` from 0 to 100.

		* **yout** : _ndarray_ <br />
		A 1darray with y-coordinates - either the array set by `setAxes()`,
		or an array of length `data.shape[0]` from 0 to 100.

		* **dout** : _ndarray_ <br />
		A 2darray with the data to be plotted. If you have set a window using
		either `setAxes()` or `setWindow()`, the data will be windowed.
		"""
		if not self.hasAxes and not self.hasWindow:
			self.x = np.linspace(0,100,data.shape[1])
			self.y = np.linspace(0,100,data.shape[0])
			self.argxmin, self.argxmax = 0, -1
			self.argymin, self.argymax = 0, -1
		elif self.hasAxes and not self.hasWindow:
			self.argxmin, self.argxmax = 0, -1
			self.argymin, self.argymax = 0, -1
		elif not self.hasAxes and self.hasWindow:
			self.x = np.linspace(0,100,data.shape[1])
			self.y = np.linspace(0,100,data.shape[0])
			self.argxmin = np.argmin(np.abs(self.x - self.xmin))
			self.argxmax = np.argmin(np.abs(self.x - self.xmax))
			self.argymin = np.argmin(np.abs(self.y - self.ymin))
			self.argymax = np.argmin(np.abs(self.y - self.ymax))
		elif self.hasAxes and self.hasWindow:
			self.argxmin = np.argmin(np.abs(self.x - self.xmin))
			self.argxmax = np.argmin(np.abs(self.x - self.xmax))
			self.argymin = np.argmin(np.abs(self.y - self.ymin))
			self.argymax = np.argmin(np.abs(self.y - self.ymax))

		xout = self.x[self.argxmin:self.argxmax:step]
		yout = self.y[self.argymin:self.argymax:step]
		dout = data[self.argymin:self.argymax:step, self.argxmin:self.argxmax:step]
		self.extent = [self.x[self.argxmin], self.x[self.argxmax], self.y[self.argymin], self.y[self.argymax]]
		return(xout, yout, dout)

	def setAxes(self, x, y, window=None):
		"""Sets the x and y axes of the singleAx object, and can apply a window.

		Note that this can be used before or after `setWindow()` - whichever
		was called last, will be used in plotting.

		**Parameters**

		* **x** : _ndarray_ <br />
		The x-coordinates. Should be 1-dimensional.

		* **y** : _ndarray_ <br />
		The y-coordinates. Should be 1-dimensional.

		* **window** : _array-like, optional_ <br />
		Format: `window = [xmin, xmax, ymin, ymax]`. Note that these are the x
		and y values, rather than their indices.

		**Returns**

		* **self** : _singleAx_
		"""
		self.hasAxes = True
		self.x, self.y = x, y
		if not window is None:
			self.hasWindow = True
			self.xmin = window[0]
			self.xmax = window[1]
			self.ymin = window[2]
			self.ymax = window[3]
		return(self)

	def setWindow(self, window=(0,100,0,100)):
		"""Applies a window to the singleAx object.

		Note that this can be used before or after `setAxes()` - whichever
		was called last, will be used in plotting.

		**Parameters**

		* **window** : _array-like, optional_ <br />
		Format: `window = [xmin, xmax, ymin, ymax]`. Note that these are the x
		and y values, rather than their indices.

		**Returns**

		* **self** : _singleAx_
		"""
		self.hasWindow = True
		self.xmin = window[0]
		self.xmax = window[1]
		self.ymin = window[2]
		self.ymax = window[3]
		return(self)

	def set_title(self, title='', **kwargs):
		"""Sets the title of the plot.

		**Parameters**

		* **title** : _string_ <br />
		The plot title.

		* ****kwargs** <br />
		All other kwargs are passed on to `matplotlib.axes.Axes.set_title`.

		**Returns**

		* **self** : _singleAx_
		"""
		self.ax.set_title(title, **kwargs)
		return(self)

	def set_xlabel(self, xlabel='', **kwargs):
		"""Sets the xlabel of the plot.

		**Parameters*

		* **xlabel** : _string_ <br />
		The xlabel.

		* ****kwargs** <br />
		All other kwargs are passed on to `matplotlib.axes.Axes.set_xlabel`.

		**Returns**

		* **self** : _singleAx_
		"""
		self.ax.set_xlabel(xlabel, **kwargs)
		return(self)

	def set_ylabel(self, ylabel='', **kwargs):
		"""Sets the ylabel of the plot.

		**Parameters**

		* **ylabel** : _string_ <br />
		The ylabel.

		* ****kwargs** <br />
		All other kwargs are passed on to `matplotlib.axes.Axes.set_ylabel`.

		**Returns**

		* **self** : _singleAx_
		"""
		self.ax.set_ylabel(ylabel, **kwargs)
		return(self)

	def set_xytitle(self, xlabel='', ylabel='', title='', **kwargs):
		"""Set the xlabel, ylabel, and title at the same time.

		Sets all three even if not all are given. Whatever you input will be applied to all three.

		For individual control, use `singleAx.set_xlabel`, `singleAx.set_ylabel`,
		or `singleAx.set_title`.

		**Parameters**

		* **ylabel** : _string_ <br />
		The ylabel.

		* **xlabel** : _string_ <br />
		The xlabel.

		* **title** : _string_ <br />
		The plot title.

		* ****kwargs** <br />
		All other kwargs are passed on
		to `matplotlib.axes.Axes.set_xlabel`, `matplotlib.axes.Axes.set_ylabel`,
		and `matplotlib.axes.Axes.set_title`.

		**Returns**

		* **self** : _singleAx_
		"""
		self.ax.set_xlabel(xlabel, **kwargs)
		self.ax.set_ylabel(ylabel, **kwargs)
		self.ax.set_title(title, **kwargs)
		return(self)

	def imshow(self, data, step=1, **kwargs):
		"""Imshows the (windowed) data. Sets origin to lower, sets the extent.

		**Parameters**

		* **data** : _ndarray_ <br />
		The data to be shown. Use the un-windowed data - the window will be
		applied automatically, if you set one.

		* **step** : _int_ <br />
		data will be returned as `data[::step,::step]` - particularly useful for
		quiver plots. <br />
		Default is `step = 1`.

		* ****kwargs** <br />
		All other kwargs are passed on to `matplotlib.axes.Axes.imshow`.

		**Returns**

		* **self** : _singleAx_
		"""
		x, y, data = self.prePlot(data, step)
		imshowargs = {'extent': self.extent, 'origin': 'lower'}
		imshowargs.update(kwargs)
		self.ax.imshow(data, **imshowargs)
		return(self)

	def quiver(self, data, step=1, rgba=False, **kwargs):
		"""Shows a quiver plot of complex data.

		**Parameters**

		* **data** : _ndarray_ <br />
		The data to be shown. Use the un-windowed data - the window will be
		applied automatically, if you set one.

		* **step** : _int_ <br />
		data will be returned as `data[::step,::step]` - particularly useful for
		quiver plots. <br />
		Default is `step = 1`.

		* **rgba** : _bool_ <br />
		If True, arrow color will correspond to the complex angle of the data. <br />
		Default is `rgba = False`.

		* ****kwargs** <br />
		All other kwargs are passed on to `matplotlib.axes.Axes.quiver`.

		**Returns**

		* **self** : _singleAx_
		"""
		xr, yr, data = self.prePlot(data, step)
		qargs = {'cmap' : cielab_cmap()}
		qargs.update(kwargs)
		self.ax.set_aspect('equal')
		if rgba:
			self.ax.set_facecolor('black')
			self.ax.quiver(xr, yr, np.real(data), np.imag(data), np.angle(data), **qargs)
		else:
			self.ax.quiver(xr,yr,np.real(data),np.imag(data),**kwargs)
		return(self)

	def rgba(self, data, step=1, brightness='intensity', alpha='uniform', cmap='uniform', **kwargs):
		"""Shows an rgba interpretation of complex data.

		**Parameters**

		* **data** : _complex ndarray_ <br />
		An array with the data to represent. Dtype may be complex or real - if real,
		the color will be uniform, and values will be represented by brightness.

		* **step** : _int_ <br />
		data will be returned as `data[::step,::step]` - particularly useful for
		quiver plots. <br />
		Default is `step = 1`.

		* **cmap** : _string, optional_ <br />
		If `cmap = 'uniform'`, the CIELAB color space will be used. Otherwise, any
		pyplot ScalarMappable may be used. <br />
		Default is `cmap = 'uniform'`.

		* **brightness** : _string, optional_ <br />
		Allowed values: `'intensity'`, `'amplitude'`, `'uniform'`. <br />
		Default is `brightness = 'intensity'`.

		* **alpha** : _string, optional_ <br />
		Allowed values: `'intensity'`, `'amplitude'`, `'uniform'`. Determines the alpha
		component of the rgba value. <br />
		Default is `alpha = 'uniform'`.

		* ****kwargs** <br />
		All other kwargs are passed on to `matplotlib.axes.Axes.imshow`.

		**Returns**

		* **self** : _singleAx_
		"""

		x, y, data = self.prePlot(data)
		data = rgba(data,brightness=brightness,alpha=alpha,cmap=cmap)
		imshowargs = {'extent': self.extent, 'origin': 'lower'}
		imshowargs.update(kwargs)
		self.ax.imshow(data, **imshowargs)
		return(self)

	def inset(self, window, **kwargs):
		"""Plots a square box with vertices defined by window.

		Default color is white.

		**Parameters**

		* **window** : _array-like_ <br />
		Format: `window = [xmin, xmax, ymin, ymax]`. Note that these are the x
		and y values, rather than their indices.

		* ****kwargs** <br />
		All other kwargs are passed on to `matplotlib.axes.Axes.plot`.

		**Returns**

		* **self** : _singleAx_
		"""
		plotargs = {'color': 'white'}
		plotargs.update(kwargs)
		self.ax.plot(np.linspace(window[0], window[1], 100),
						np.zeros(100) + window[2], **plotargs)
		self.ax.plot(np.linspace(window[0], window[1],100),
						np.zeros(100)+window[3], **plotargs)
		self.ax.plot(np.zeros(100) + window[0],
						np.linspace(window[2], window[3], 100), **plotargs)
		self.ax.plot(np.zeros(100) + window[1],
						np.linspace(window[2], window[3], 100),
						**plotargs)
		return(self)

def subplots(rc=11, **kwargs):
		"""Creates a fig, ax instance but replaces ax with singleAx.

		Behaves almost identically to matplotlib.pyplot.subplots(), but replaces each
		`matplotlib.axes.Axes` object with a `wsp_tools.pyplotwrapper.singleAx`
		object.

		Each `wsp_tools.pyplotwrapper.singleAx` object in turn behaves just like a
		normal `Axes` object, but with added methods.

		**Parameters**

		* **rc** : _int_ <br />
		First digit - nrows. Second digit - ncols. <br />
		Default is `rc = 11`.

		* **squeeze** : _bool_ <br />
		If true, extra dimensions are squeezed out from the returned array of Axes. <br />
		Default is `squeeze = False`.

		* ****kwargs** <br />
		All other kwargs are passed on to `matplotlib.axes.Axes.subplots`.

		**Returns**

		* **fig** : _Figure_ <br />

		* **ax** : _singleAx_ or array of _singleAx_ objects <br />
		"""
		subplotsargs = {'tight_layout': True, 'squeeze': False}
		subplotsargs.update(kwargs)
		fig, ax = plt.subplots(nrows=rc//10, ncols=rc%10, **subplotsargs)
		for i in range(ax.shape[0]):
			for j in range(ax.shape[1]):
				ax[i][j] = singleAx(ax[i][j])
		return(fig, np.array(ax))
